from typing import List


def divide_path(path: str, separator='/', divider='|') -> List[str]:
    """ Dividing Single Path to Multiple Path

    ['a/b/c', 'a/b/d']
    >>> divide_path('a/b/c|d')

    ['a/1/b','a/2/b']
    >>> divide_path('a/1|2/b')
    """
    segments = path.split(separator)
    result = []
    for segment in segments:
        divided_segments = segment.split(divider)

        if len(result) > 0:
            divided_paths = []
            for divided_segment in divided_segments:
                for divided_path in result:
                    divided_paths.append(
                        divided_path + separator + divided_segment
                    )
            result = divided_paths
        else:
            for divided_segment in divided_segments:
                result.append(divided_segment)

    return result



import re

_NUMERIC = r'([+-])?(\d+(?:\.\d+)?)'


def casting(target):
    """ Auto-casting string as it cast-able

    - bool
        * "true" and "false" for Normal
        * "True" and "False" for `Python`
        * "TRUE" and "FALSE" for `Database`
    - None
        * "null" for Normal
        " "undefined" for `Javascript`
        * "NULL" for `Database`
    - int
        * Number without Decimal Precision Mark
    - float
        * Number with Decimal Precision Mark

    :param target: Cast-able String
    :return: Auto-casted target
    """

    if not isinstance(target, str):
        return target

    if target == 'true' or target == 'True' or target == 'TRUE':
        return True

    if target == 'false' or target == 'False' or target == 'FALSE':
        return False

    if target == 'null' or target == 'NULL' or target == 'undefined':
        return None

    matched = re.fullmatch(_NUMERIC, target)
    if matched:
        if '.' in target:
            return float(target)

        return int(target)

    return target


def to_fixed(value: float, precision: int):
    return float(('%0.' + str(precision) + 'f') % value)



import os
from os.path import dirname, isfile, join
from sys import path


def insert_path_by_identified_file(identified_file: str):
    root = os.getcwd()

    while root and not isfile(join(root, identified_file)):
        root = dirname(root)
        print(root)

    if not root:
        raise IOError(f'Cannot found ${identified_file}')

    path.insert(0, root)



from datetime import datetime

from sqlalchemy import or_, Column, and_
from sqlalchemy.sql import sqltypes
from sqlalchemy.sql.sqltypes import String, TypeDecorator

from py_dot.core.date import Period
from py_dot.core.route import from_query_string


def _get_period_logic(column: Column, period: Period):
    begin = period.begin
    end = period.end

    if begin and end:
        # return f'{column.name}.between({begin}, {end})'
        print(begin, end)
        return column.between(begin, end)

    if begin:
        # return f'{column.name} >= {begin}'
        return column >= begin

    if end:
        # return f'{column.name} <= {end}'
        return column <= end

    raise ValueError('Period has not Begin and End Time')


def _get_normal_logic(column: Column, value):
    column_type = column.type

    if isinstance(column_type, sqltypes.DateTime):
        if not isinstance(value, (Period, datetime)):
            raise ValueError(f'`{column.name}` Keyword should be cast-able to Timestamp')

        if isinstance(value, Period):
            return _get_period_logic(column, value)

    if isinstance(value, (bool, datetime)):
        return column == value

    if _is_string_type(column):
        value = str(value)

        if _is_quote(value):
            return column.like(value[1:-1])

        value = value.replace('%', '\\%')
        # return f'{column.name}.like(%{value}%)'
        return column.like(f'%{value}%')

    # return f'{column.name} == {value[1:-1]}'
    return column == value


def _is_quote(value: str):
    if value.startswith('"') and value.endswith('"'):
        return True

    if value.startswith("'") and value.endswith("'"):
        return True

    return False


def _is_string_type(column: Column):
    if not hasattr(column, 'dot_is_string_type_'):
        setattr(column, 'dot_is_string_type_', _is_string_type_test(column))

    return column.dot_is_string_type_


def _is_string_type_test(column: Column):
    """
    todo: Make to Global Function
    """
    column_type = column.type

    if isinstance(column_type, String) or String in column_type.__class__.__mro__:
        return True

    if isinstance(column_type, TypeDecorator):
        impl = getattr(column_type.__class__, 'impl')
        if isinstance(impl, String) or String in impl.__class__.__mro__:
            return True

    return False


def filter_from_string(
    model: object,
    query_string: str,
    use_snake_case=False,
    use_default_timezone=False
):
    """ Make Sqlalchemy Query Filter by Query String

    :param model: ORM Class
    :param query_string: Query the Search String
    :param use_snake_case: Using snake_case to key
    :param use_default_timezone:
    :raise ValueError: when target condition is not a Column of Model
    """

    if not hasattr(model, '__table__'):
        raise ValueError('Not a Model Class')

    columns = model.__table__.columns

    conditions = from_query_string(
        query_string,
        use_snake_case=use_snake_case,
        use_default_timezone=use_default_timezone
    )

    filters = []

    for column_name in conditions:
        if not hasattr(columns, column_name):
            raise ValueError(f'{column_name} is not a Attribute')

        column: Column = getattr(columns, column_name)

        if not isinstance(column, Column):
            raise ValueError(f'{column_name} is not a Column')

        condition = conditions[column_name]

        if isinstance(condition, list):
            equal_values = []
            like_logics = []
            for or_condition in condition:
                if isinstance(or_condition, (bool, int, float)):
                    equal_values.append(or_condition)
                    continue

                if isinstance(or_condition, str) and _is_quote(or_condition):
                    equal_values.append(or_condition[1:-1])
                    continue

                like_logics.append(
                    _get_normal_logic(column, or_condition)
                )

            if equal_values and like_logics:
                # filters.append(
                #     '(' +
                #     ' OR '.join(
                #         [
                #             f'{column_name}.in({",".join(map(str, equal_values))})',
                #             *like_values
                #         ]
                #     )
                #     + ')'
                # )
                filters.append(
                    or_(
                        column.in_(*equal_values),
                        *like_logics
                    )
                )
                continue

            if equal_values:
                # filters.append(
                #     f'{column_name}.in({",".join(map(str, equal_values))})'
                # )
                filters.append(
                    column.in_(*equal_values)
                )
                continue

            # filters.append(f'({" OR ".join(like_values)})')
            filters.append(*like_logics)
            continue

        filters.append(
            _get_normal_logic(column, condition)
        )

    # return ' AND \n'.join(filters)
    return and_(*filters)

# query_string_parts = [
#     'single_int:1',
#     'single_str:foo',
#     'single_eq_str:"foo"',
#     'single_begin:20211130~',
#     'single_end:~20211130',
#     'single_period:20211129~20211130',
#
#     'multi_int:1,2',
#     'multi_str:foo,bar',
#     'multi_mix:1,foo',
#     'multi_eq_mix:1,"foo"',
#     'multi_eq_mix_with_numeric_str:"1","foo"'
# ]
# filters = filter_from_query_string({}, ';'.join(query_string_parts))
# print(filters)



import unittest
from datetime import datetime

from py_dot.core.date import Period
from py_dot.core.route import from_query_string
from py_dot.core.route.route import divide_path


class TestRoute(unittest.TestCase):
    def test_from_query_string(self):
        self.assertDictEqual(
            from_query_string('foo:bar'),
            {
                'foo': 'bar'
            }
        )

        self.assertDictEqual(
            from_query_string('foo:bar,baz'),
            {
                'foo': ['bar', 'baz']
            }
        )

        self.assertEqual(
            from_query_string('foo:bar;baz:qux'),
            {
                'foo': 'bar',
                'baz': 'qux'
            }
        )

        self.assertIsInstance(
            from_query_string('foo:20211130')['foo'],
            datetime
        )

        self.assertIsInstance(
            from_query_string('foo:20211130~21211201')['foo'],
            Period
        )

    def test_divide_path(self):
        self.assertListEqual(
            divide_path('a/b/c|d'),
            ['a/b/c', 'a/b/d']
        )

        result = divide_path('a|b/c|d')
        result.sort()
        expect = ['a/c', 'a/d', 'b/c', 'b/d']
        expect.sort()

        self.assertListEqual(result, expect)

        self.assertListEqual(
            divide_path('a/1|2/b'),
            ['a/1/b', 'a/2/b']
        )



import base64
import re


def to_snake(value: str) -> str:
    """ String to snake_case

    - camelCase: camel_case
    - space case: space_case
    - kebab-case: kebab_case

    :param value: Target Value
    :return:
    """
    words = re.findall(r'([A-Z]?[^_\-\sA-Z]+)', value)
    return '_'.join(map(lambda x: x.lower(), words))


def to_base64(value: str):
    encoded_value = base64.urlsafe_b64encode(value.encode('utf-8'))
    return str(encoded_value, 'utf-8')


def from_base64(value: str):
    decoded_value = base64.urlsafe_b64decode(value)
    return str(decoded_value, 'utf-8')


from dataclasses import dataclass
from datetime import datetime
from math import ceil
from typing import Type, Union, List, Tuple, Dict, TypedDict, Optional

from py_dot.core.date import Period, DateParts


class SummaryTimeUnit:
    year = 'year'
    month = 'month'
    date = 'date'
    hour = 'hour'
    minute = 'minute'
    second = 'second'

    def __init__(
            self,
            year=None,
            month=None,
            date=None,
            hour=None,
            minute=None,
            second=None,
            default=None
    ):
        if year is None and month is None and date is None and hour is None and minute is None and second:
            if default is None:
                raise ValueError('One of time value is required at least.')

        self.year = year
        self.month = month
        self.date = date
        self.hour = hour
        self.minute = minute
        self.second = second
        self.default = default


def get_summary_unit(
        period: Period,
        unit: Union[SummaryTimeUnit, Type[SummaryTimeUnit]] = SummaryTimeUnit,
        year_days: int = 365,
        month_days: int = 30,
        year_from: int = 3,
        month_from: int = 6,
        date_from: int = 3,
        hour_from: int = 28,
        minute_from: int = 3599
) -> Tuple[any, int, int]:
    duration_time = period.duration_time

    begin = DataSummaryTimeDigits(period.begin)
    end = DataSummaryTimeDigits(period.end)

    days = ceil(duration_time / 86400)

    if unit.year is not None:
        if days >= year_days * year_from:
            return unit.year, begin.y, end.y

    if unit.month is not None:
        if days >= month_days * month_from:
            return unit.month, begin.ym, end.ym

    if unit.date is not None:
        if days >= date_from:
            return unit.date, begin.ymd, end.ymd

    if unit.hour is not None:
        if days >= hour_from:
            return unit.hour, begin.ymdh, end.ymdh

    if unit.minute is not None:
        if duration_time >= minute_from:
            return unit.minute, begin.ymdhi, end.ymdhi

    return unit.default, period.begin, period.end


class DataSummaryTimeDigits:
    y: int
    ym: int
    ymd: int
    ymdh: int
    ymdhi: int

    def __init__(self, time: datetime = None, parts: DateParts = None):
        if parts is None:
            if time is None:
                raise ValueError('time or parts Arguments is Required.')

            parts = DateParts(time)

        y = parts.y
        m = parts.m
        d = parts.d
        h = parts.h
        i = parts.i

        ym = y + m
        ymd = ym + d
        ymdh = ymd + h
        ymdhi = ymdh + i

        self.y = int(y)
        self.ym = int(ym)
        self.ymd = int(ymd)
        self.ymdh = int(ymdh)
        self.ymdhi = int(ymdhi)

    def get(self, unit: str):
        pass


@dataclass
class DataSummaryTimeStrings:
    y: str
    ym: str
    ymd: str
    ymdh: str
    ymdhi: str

    def __init__(self, time: datetime = None, parts: DateParts = None):
        if parts is None:
            if time is None:
                raise ValueError('time or parts Arguments is Required.')

            parts = DateParts(time)

        y, m, d, h, i = parts

        ym = y + '-' + m
        ymd = ym + '-' + d
        ymdh = ymd + ' ' + h
        ymdhi = ymdh + ':' + i

        self.y = y
        self.ym = ym
        self.ymd = ymd
        self.ymdh = ymdh
        self.ymdhi = ymdhi


class DataSummaryTimes:
    digits: DataSummaryTimeDigits
    strings: DataSummaryTimeStrings

    def __init__(self, time: datetime):
        parts = DateParts(time)

        self.digits = DataSummaryTimeDigits(parts=parts)
        self.strings = DataSummaryTimeStrings(parts=parts)


def iterator_to_kdv(
        items: List[Union[Tuple, List]],
        key_index=0,
        date_index=1,
        value_index=2,
        header: Tuple[str, str] = None
) -> Dict[str, List[Tuple[str, Union[int, float, str]]]]:
    """ Get Key-date-value Dict

    :param items: Iterator
    :param key_index: Index of items to make key
    :param date_index: Index of items to make date
    :param value_index: Index of items to make value
    :param header: First Item to using as heading
    """
    result = {}

    for item in items:
        key = item[key_index]
        date = item[date_index]
        value = item[value_index]

        item = (date, value)
        if key in result:
            result[key].append(item)
        else:
            if header:
                result[key] = [header, item]
                continue
            result[key] = [item]

    return result


class _KdvWithSummary(TypedDict):
    count: int
    sum: int
    average: Optional[float]
    min: int
    max: int
    series: List[Tuple[str, Union[int, float, str]]]


def iterator_to_kdv_with_summary(
        items: List[Union[Tuple, List]],
        key_index=0,
        date_index=1,
        value_index=2,
        header: Tuple[str, str] = None
) -> Dict[str, _KdvWithSummary]:
    """ Get Key-date-value Dict with Summaries
    :param items: Iterator
    :param key_index: Index of items to make key
    :param date_index: Index of items to make date
    :param value_index: Index of items to make value
    :param header: First Item to using as heading
    """
    result = {}

    for item in items:
        key = item[key_index]
        date = item[date_index]
        value = item[value_index]

        item = (date, value)
        if key in result:
            summary = result[key]
            summary['count'] = summary['count'] + 1
            summary['sum'] = summary['sum'] + value
            summary['min'] = min(summary['min'], value)
            summary['max'] = max(summary['max'], value)
            summary['series'].append(item)
        else:
            result[key] = _KdvWithSummary(
                count=1,
                sum=value,
                average=value,
                min=value,
                max=value,
                series=[item] if header is None else [header, item]
            )

    for key in result:
        summary = result[key]
        summary['average'] = summary['sum'] / summary['count']

    return result



import unittest

from py_dot.core.date import Period
from py_dot.data import get_summary_unit


class TestDataCore(unittest.TestCase):
    def test_get_summary_unit(self):

        self.assertEqual(
            get_summary_unit(
                Period(
                    '20000101000000',
                    '20000101000100'
                )
            ),
            None
        )

        self.assertEqual(
            get_summary_unit(
                Period(
                    '20000101000000',
                    '20000101005959'
                )
            ),
            'minute'
        )


from enum import Enum
from typing import Any, TypeVar, Dict, Union

from sqlalchemy import text
from sqlalchemy.orm.attributes import History


class Change:
    """ Simplified `History` Structure

    """
    previous_value: Any
    current_value: Any

    def __init__(self, history: History):
        changed = history.has_changes()
        self.changed = changed

        if not changed:
            return

        self.previous_value = history.deleted[0]
        self.current_value = history.added[0]


Changes = Dict[str, Change]


def defaults(value) -> Dict:
    """ Return Column `default` and `server_default` as dict

    Column(String, default='default_value', server_default=text('"default_value"')

    >>> Column(String, **defaults('default_value'))
    """

    if isinstance(value, Enum):
        value = value.value

    return {
        'default': value,
        'server_default': text(
            str(value) if not isinstance(value, str) else f"'{value}'"
        )
    }



import math
import re


class ArrangeCalculator:
    size: int
    max_depth: int
    max_depth_length: int
    child_suffix: str

    def __init__(self, size: int = 3, length: int = 19):
        max_depth = math.floor(length / 3) - 1
        max_depth_length = max_depth * 3
        child_suffix = ''.ljust(max_depth_length, '0')

        self.size = size
        self.length = length
        self.max_depth = max_depth
        self.max_depth_length = max_depth_length
        self.child_suffix = child_suffix

    def item(self, arrange: int):
        return ArrangeItemCalculator(self, arrange)

    def after_item(self, arrange: int):
        return ArrangeItemCalculator(self, self.item(arrange).after)

    def reviser(self, arrange: int):
        return ArrangeReviser(self, arrange)


class ArrangeItemCalculator:
    calculator: ArrangeCalculator
    arrange: int
    root_order: str
    depth: int
    child_depth: int

    _child_order_mount: int
    _first_child: int
    _last_child: int
    _sibling_order_amount: int

    @property
    def child_order_amount(self):
        if hasattr(self, '_child_order_amount'):
            return self._child_order_amount

        calculator = self.calculator
        depth = self.depth

        child_order_amount = int(
            '1' + ''.ljust(
                calculator.size * (calculator.max_depth - depth - 1),
                '0'
            )
        )

        self._child_order_mount = child_order_amount
        return child_order_amount

    @property
    def first_child(self):
        if hasattr(self, '_first_child'):
            return self._first_child

        first_child = self.arrange + self.child_order_amount
        self._first_child = first_child

        return first_child

    @property
    def last_child(self):
        if hasattr(self, '_last_child'):
            return self._last_child

        last_child = self.after - self.child_order_amount
        self._last_child = last_child

        return last_child

    @property
    def sibling_order_amount(self):
        if hasattr(self, '_sibling_order_amount'):
            return self._sibling_order_amount

        sibling_order_amount = int(
            str(self.child_order_amount) + ''.rjust(self.calculator.size, '0')
        )

        self._sibling_order_amount = sibling_order_amount
        return sibling_order_amount

    @property
    def after(self):
        return self.arrange + self.sibling_order_amount

    def __init__(self, calculator: ArrangeCalculator, arrange: int):

        if arrange is None:
            print("Arrange is None!!!!!!!!!!!!!!")
            arrange = 0

        string_arrange = str(arrange)
        root_order = string_arrange[0:len(string_arrange) - calculator.max_depth_length]

        depth = math.ceil(
            len(re.sub('0+$', '', string_arrange[len(root_order):]))
            / calculator.size
        )

        self.calculator = calculator
        self.arrange = arrange
        self.root_order = root_order
        self.depth = depth
        self.child_depth = depth + 1


class ArrangeReviser:
    calculator: ArrangeCalculator
    arrange: int

    def __init__(self, calculator: ArrangeCalculator, arrange: int):
        self.calculator = calculator
        self.arrange = arrange



from typing import Dict, Any

from py_dot.core.strings import to_snake


def merge(
        target: Dict,
        source: Dict,
        list_merge=True,
        snake_case=False
) -> Dict:
    """ Merge Two Dict to Single Dict

    >>> merge({'a': [1], 'b': True}, {'a':[2], 'c': False})
    ... {
    ... 'a': [1, 2]
    ... 'b': True,
    ... 'c': False
    ... }
    """
    result = {**target}

    if snake_case:
        for name in source:
            snake_name = to_snake(name)
            source_value = source[name]
            if name not in result:
                result[snake_name] = source_value
                continue

            target_value = result[name]

            if isinstance(source_value, dict) and isinstance(target_value, dict):
                result[snake_name] = merge(target_value, source_value, list_merge, snake_case)
                continue

            if list_merge and isinstance(source_value, list) and isinstance(target_value, list):
                result[snake_name] = [*target_value, *source_value]
                continue

            result[name] = source_value
    else:
        for name in source:
            source_value = source[name]
            if name not in result:
                result[name] = source_value
                continue

            target_value = result[name]

            if isinstance(source_value, dict) and isinstance(target_value, dict):
                result[name] = merge(target_value, source_value, list_merge)
                continue

            if list_merge and isinstance(source_value, list) and isinstance(target_value, list):
                result[name] = [*target_value, *source_value]
                continue

            result[name] = source_value

    return result


def to_snake_key(values: Dict[str, Any]) -> Dict:
    snake_values = {}

    for name in values:
        snake_values[to_snake(name)] = values[name]

    return snake_values



import re
import time
from dataclasses import dataclass
from datetime import datetime, timezone, timedelta
from typing import List, Union

_PERIOD = r'^([+-]?\d+(?:\.\d+)?)?~([+-]?\d+(?:\.\d+)?)?$'

_DIGIT_TIMESTAMP = r'^([+-])?(\d{4})(\d\d)(\d\d)(?:(\d\d)(?:(\d\d)(?:(\d\d)(\d+)?)?)?)?(?:\.(\d+))?$'

now = time.time()
TIMEZONE_OFFSET_HOUR = (datetime.fromtimestamp(now) - datetime.utcfromtimestamp(now)).total_seconds() / 3600


class Period:
    begin: datetime = None
    end: datetime = None

    _duration_time: int

    @property
    def duration_time(self) -> int:
        if not hasattr(self, '_duration_time'):
            self._duration_time = int((self.end - self.begin).total_seconds())

        return self._duration_time

    def __init__(
        self,
        expression: str = None,
        matched: re.Match = None,
        begin: Union[str, datetime] = None,
        end: Union[str, datetime] = None,
        use_default_timezone=None
    ):
        if not (expression or matched or begin or end):
            raise ValueError('`begin`, `end` or `period_string` is required.')

        if matched is None:
            matched = is_period_string(expression)

            if matched is None:
                raise ValueError('Invalid Period String:' + expression)

        if matched:
            begin, end = matched.groups()

        if begin:
            self.begin = from_digit_timestamp(begin, use_default_timezone=use_default_timezone)

        if end:
            self.end = from_digit_timestamp(end, use_default_timezone=use_default_timezone, to_end=True)

    def __str__(self):
        begin = self.begin
        end = self.end

        if begin and end:
            return f"{begin.__str__()}~{end.__str__()}"

        if begin:
            return begin.__str__() + '~'

        return '~' + end.__str__()

    def split(self, seconds: int) -> List['Period']:
        pass


@dataclass
class DateParts:
    y: str
    m: str
    d: str
    h: str
    i: str

    def __init__(self, time: datetime):
        self.y = str(time.year).rjust(2, '0')
        self.m = str(time.month).rjust(2, '0')
        self.d = str(time.day).rjust(2, '0')
        self.h = str(time.hour).rjust(2, '0')
        self.i = str(time.minute).rjust(2, '0')


def is_digit_timestamp(timestamp_string: str) -> re.Match:
    """ Check the String be Digit Timestamp-able String

    :param timestamp_string:
    """
    return re.fullmatch(_DIGIT_TIMESTAMP, timestamp_string)


def _get_part(value: int, default: int):
    return value if value else default


def from_digit_timestamp(
    timestamp_string: str,
    matched: re.Match = -1,
    use_default_timezone=False,
    to_end=False
) -> datetime:
    """ Get Datetime from Digit Timestamp String

    UTC:

    - 12341212
    - 1234121213
    - 123412121212
    - 12341212121212
    - 12341212121212123

    +12:

    - 12341212121212123.12

    -12:

    - -12341212121212123.12

    :param timestamp_string:
    :param matched:
    :param use_default_timezone: Using Default Timezone when has not Timezone Digit
    """
    if matched == -1:
        matched = is_digit_timestamp(timestamp_string)

    if matched is None:
        raise ValueError(f'Invalid Digit Timestamp String: {timestamp_string}')

    flag, y, m, d, h, i, s, u, z = matched.groups()

    timezone_hours = 0
    if z:
        timezone_hours = int(z) if flag != '-' else -int(z)
    elif use_default_timezone:
        timezone_hours = TIMEZONE_OFFSET_HOUR

    return datetime(
        int(y),
        int(m),
        int(d),
        int(_get_part(h, 23 if to_end else 0)),
        int(_get_part(i, 59 if to_end else 0)),
        int(_get_part(s, 59 if to_end else 0)),
        int(_get_part(u, 999999 if to_end else 0)),
        tzinfo=timezone(
            timedelta(
                hours=timezone_hours
            )
        )
    )


def is_period_string(period_string: str) -> Union[re.Match, None]:
    """ Check the String be Period-able String
    """
    if period_string != '~':
        return re.fullmatch(_PERIOD, period_string)

    return None

# def from_period_string(period_string: str, matched: re.Match = -1) -> Period:
#     """ Get Begin and End Date from Period Expression String
#
#     - [DIGIT_TIMESTAMP]~[DIGIT_TIMESTAMP]
#     - [DIGIT_TIMESTAMP]~
#     - ~[DIGIT_TIMESTAMP]
#
#     :param period_string:
#     :param matched:
#     """
#     if matched == -1:
#         matched = is_period_string(period_string)
#
#     if matched is None:
#         raise ValueError(f'Invalid Period String: {period_string}')
#
#     return Period(matched[0], matched[1])



from functools import wraps

def outer_of(inner_decorator):
    """ Create Decorator as Outer Wrapped of argument

    >>> def foo(f):
    ...    def d():
    ...        print('inner')
    ...        return f()
    ...
    ...    return d
    ...
    ...
    ...@outer_of(foo)
    ...def bar(f):
    ...    def d():
    ...        print('outer')
    ...        return f()
    ...
    ...    return d
    ...
    ...@bar
    ...def bar_and_foo():
    ...    pass

    :param inner_decorator: Inner Decorator of Outer Decorator
    :return:
    """

    def outer_wrap(outer_decorator):
        def inner_wrap(callback):
            child_decorated = outer_decorator(inner_decorator(callback))

            @wraps(callback)
            def decorated_wrap(*args, **child_kwargs):
                return child_decorated(*args, **child_kwargs)

            return decorated_wrap

        return inner_wrap

    return outer_wrap



from threading import Thread, Event
from time import sleep
from typing import Callable


class TimerEvent(Event):
    """Timer Operator
    """

    def __init__(self):
        self._event = Event()

    def clear(self):
        """ Stop the Timer"""
        self._event.set()


def set_timeout(callback: Callable, delay: int, daemon=False) -> TimerEvent:
    """ Set Timer to Run after delay

    Print after 3s:
    >>> def foo(timer_: TimerEvent):
    ...     print('foo')
    ... timer = set_timeout(foo, 3)


    Disable Timer:
    >>> timer.clear()
    """

    def set_timeout_thread(event_: TimerEvent):
        sleep(delay)
        if not event_.is_set():
            callback()

    event = TimerEvent()
    thread = Thread(target=set_timeout_thread, args=(event,), daemon=daemon)
    thread.start()

    return event


def set_interval(callback: Callable, delay: int, daemon=False):
    """ Set Timer to Run per delay

    Print per 3s:
    >>> def foo(timer_: TimerEvent, index: int):
    ...     print('foo')
    ... timer = set_interval(foo, 3)


    Disable Timer:
    >>> timer.clear()
    """

    def set_interval_thread(event_: TimerEvent):
        index = 0
        while not event_.is_set():
            sleep(delay)
            callback(event, index)
            index += 1

    event = TimerEvent()
    thread = Thread(target=set_interval_thread, args=(event,), daemon=daemon)
    thread.start()

    return thread


def timeout(delay: int, daemon=False):
    """ set_timeout sugar decorator

    Print after 3s
    >>> @timeout(3)
    ... def foo(timer: TimerEvent):
    ...     print('foo')

    Disable Timer:
    >>> foo.clear()
    """

    def set_timeout_decorator(callback: Callable):
        return set_timeout(callback, delay, daemon)

    return set_timeout_decorator


def interval(delay: int, daemon=False):
    """ set_interval sugar decorator

    Print per 3s
    >>> @interval(3)
    ... def foo(timer: TimerEvent, index: int):
    ...     print('foo')

    Disable Timer:
    >>> foo.clear()
    """

    def set_interval_decorator(callback: Callable):
        return set_interval(callback, delay, daemon=daemon)

    return set_interval_decorator



import unittest

from py_dot.core import casting


class TestCore(unittest.TestCase):
    def test_casting(self):
        self.assertEqual(casting('true'), True)
        self.assertEqual(casting('True'), True)
        self.assertEqual(casting('TRUE'), True)
        self.assertEqual(casting('tRUe'), 'tRUe')
        self.assertEqual(casting('false'), False)
        self.assertEqual(casting('False'), False)
        self.assertEqual(casting('FALSE'), False)
        self.assertEqual(casting('fALSe'), 'fALSe')
        self.assertEqual(casting('null'), None)
        self.assertEqual(casting('NULL'), None)
        self.assertEqual(casting('undefined'), None)
        self.assertEqual(casting('Nope'), 'Nope')
        self.assertEqual(casting('1'), 1)
        self.assertEqual(casting('1.1'), 1.1)
        self.assertEqual(casting(True), True)
        self.assertEqual(casting(False), False)
        self.assertEqual(casting(None), None)








from typing import Dict, Any

from py_dot.core import casting
from py_dot.core.date import is_period_string, Period, is_digit_timestamp, from_digit_timestamp
from py_dot.core.dicts import to_snake_key

FILTER_KEYWORD_SPACE = '+'
FILTER_KEYWORD_OR = ','


def _get_filter_keyword(
    value: str,
    keyword_space=FILTER_KEYWORD_SPACE,
    keyword_or=FILTER_KEYWORD_OR,
    use_default_timezone=None
):
    or_values = value.split(keyword_or)
    result = []

    for or_value in or_values:
        if len(or_value) == 0:
            continue

        matched_period = is_period_string(or_value)
        if matched_period:
            result.append(Period(matched=matched_period, use_default_timezone=use_default_timezone))
            continue

        matched_digit_timestamp = is_digit_timestamp(or_value)

        if matched_digit_timestamp:
            result.append(from_digit_timestamp(or_value, matched=matched_digit_timestamp))
            continue

        or_value = casting(or_value)
        if isinstance(or_value, str):
            or_value = or_value.replace(keyword_space, ' ')

        result.append(or_value)

    return result if len(result) > 1 else result[0]


def from_query_string(
    query_string: str,
    use_snake_case=False,
    use_default_timezone=False
) -> Dict[str, Any]:
    """ Get Query Condition from Query String

    `foo:bar`:
    >>> {
    ... 'foo': 'bar'
    ...}

    `foo:bar,baz`:
    >>>{
    ...    'foo': ['bar', 'baz']
    ...}


    foo:bar;baz:qux:
    >>>{
    ...    'foo': 'bar',
    ...    'baz': 'qux'
    ...}

    :param query_string: Shortened Search Condition
    :param use_snake_case: Using snake_case to key
    :param use_default_timezone:
    :return:
    """

    conditions = {}
    end = len(query_string)
    stack = ''

    key = ''
    current = 0
    get_column = True

    while current < end:
        char = query_string[current]
        current += 1

        if get_column:
            if char == ':':
                key = stack
                stack = ''
                get_column = False
                continue

        else:
            if char == ';':
                conditions[key] = _get_filter_keyword(stack, use_default_timezone=use_default_timezone)
                stack = ''
                get_column = True
                continue

        stack += char

    if key and stack:
        conditions[key] = _get_filter_keyword(stack, use_default_timezone=use_default_timezone)

    if use_snake_case:
        return to_snake_key(conditions)

    return conditions



import unittest

from py_dot.core.dicts import merge


class TestDicts(unittest.TestCase):

    def test_merge_list(self):
        self.assertListEqual(
            merge({'1': [1]}, {'1': [2]})['1'],
            [1, 2]
        )

    def test_merge_dict(self):
        self.assertDictEqual(
            merge({'1': {'a': 1}}, {'1': {'b': 2}}),
            {
                '1': {
                    'a': 1,
                    'b': 2
                }
            }
        )



import pathlib

import setuptools

HERE = pathlib.Path(__file__).parent
README = (HERE / 'README.md').read_text()

"""
>>> py setup.py sdist bdist_wheel
... twine check dist/*
... twine upload dist/*
"""

setuptools.setup(
    name='py-dot',
    version='0.2.30',
    description='Python Base Development Library',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/protyliss/py-dot',
    author='Protyliss',
    author_email='protyliss@gmail.com',
    license='MIT',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8'
    ],
    packages=setuptools.find_packages(),
    include_package_data=True
    # install_requires=[]
    # entry_points={}
)



from abc import abstractmethod


class Operator:
    def __init__(self, *logic):
        self.logics = logic

    @abstractmethod
    def __bool__(self):
        pass


class Or(Operator):
    def __bool__(self):
        return any(self.logics)


class And(Operator):
    def __bool__(self):
        return all(self.logics)


class Not(Operator):
    def __bool__(self):
        return not self.logics


import getopt
import sys
from typing import List, Dict


def get_command_values(long_names: List[str]) -> Dict:
    """ Get Command Arguments as Optional

    `getopt` cannot ignore optional argument, then Use this.

    :param long_names: parameter name starts without '--'
    :return:
    """
    argv = sys.argv[1:]

    values = {}

    space_long_name = False
    for arg_value in argv:
        if arg_value[0:2] != '--':
            if space_long_name:
                if space_long_name in values:
                    value = values[space_long_name]
                    if isinstance(value, list):
                        value.append(arg_value)
                    else:
                        values[space_long_name] = [value, arg_value]
                else:
                    values[space_long_name] = arg_value
            continue

        space_long_name = False
        current = 0
        for long_name in long_names:
            if arg_value.startswith(long_name + '=', 2):
                value = arg_value[len(long_name) + 3:]
                values[long_name] = value

                if not long_names:
                    return values

                break

            if arg_value.startswith(long_name, 2):
                long_names.pop(current)
                space_long_name = long_name

                break

            current += 1

    return values



import json
from os.path import join, dirname
from typing import Dict

from py_dot.core.dicts import merge


def load(json_filename: str, extends_property: str = 'extends') -> Dict:
    """ json.loads with assign `extends` file as source

    """
    extends_value = json.loads(open(json_filename, 'r').read())
    extends_values = [extends_value]

    while extends_property in extends_value:
        json_filename = join(dirname(json_filename), extends_value[extends_property])
        extends_value = json.loads(open(json_filename, 'r').read())
        extends_values.append(extends_value)

    extends_length = len(extends_values) - 1
    while extends_length > 0:
        previous_index = extends_length - 1
        extends_values[previous_index] = merge(extends_values[extends_length], extends_values[previous_index], False)
        extends_length -= 1

    return extends_values[0]
